import os
from types import LambdaType
from typing import Dict, List, Callable

Methods = Dict[str, LambdaType]

_methods = {}


def _configure_methods(method: str) -> Callable:
    if method in _methods:
        return _methods[method]
    if 'env' == method:
        _methods.update({'env': lambda s: None if not s in os.environ else os.environ[s]})
    if 'gs' == method:
        _methods.update(_configure_secret_manager())
    return _methods[method]


def _gsecrets(secrets, coords: str) -> str:
    if coords[0] != '/':
        raise Exception(f'ERR: Unrecognized GS coordinate {coords}.')
    parts = coords[1:].split('/')
    if len(parts) not in [2, 3]:
        raise Exception(f'Need project and secret name, and optional version, for method "gs", got {coords}.')
    version = 1 if len(parts) == 2 else parts[2]
    path = f'projects/{parts[0]}/secrets/{parts[1]}/versions/{version}'
    return secrets.access_secret_version(path).payload.data.decode("utf-8")


def _configure_secret_manager() -> Methods:
    try:
        from google.cloud import secretmanager
        from google.auth.exceptions import DefaultCredentialsError
        secrets = secretmanager.SecretManagerServiceClient()
    except NameError:
        raise Exception('ERR: Requested extra "google.cloud.secretmanager" is not available.')
    return {'gs': lambda s: _gsecrets(secrets, s)}



from typing import Any, Tuple, Union, Optional

import yaml

from .methods import _configure_methods as methods


def load(config_file: str, flatten: bool = False, override: dict = None):
    """
    :param config_file: Resolvable path of the YAML file containing the config.
    :param (optional) flatten: Boolean value, truthy if you want to flatten the config file.
    :param (optional) override: Map of values to use instead of the config file.
    :param (optional) configure_methods: List of interpolation methods to use.
    """
    with open(config_file, 'r') as inimage:
        config = yaml.safe_load(inimage)
        _interpolate_leafs(config)
        if flatten:
            config = _flatten(config)
        if override:
            config.update(override)
        return config


def _interpolate_leafs(config: dict) -> None:
    for child_key in config.keys():
        _interpolate(config, child_key)


def _get_interpolated(value: str) -> str:
    if len(value) < 3:
        return value
    if '${' + (name := value[2:-1]) + '}' != value:
        return value
    parts = value.count(':')
    if parts >= 2:
        method, key, default = name.split(':', 2)
    elif parts == 1:
        method, key = name.split(':')
        default = None
    else:
        raise Exception(f'Need at least a method and name for interpolation, got {value}.')
    interpolated = methods(method)(key)
    if interpolated is None:
        return _type_interpolated(default)
    return interpolated


def _type_interpolated(value: Optional[str]) -> Optional[Union[str, int, bool, float]]:
    """ Using default value, let's look at it."""
    if value is None:
        return None
    if value.capitalize() == 'True':
        return True
    if value.capitalize() == 'False':
        return False
    if value.isdigit():
        return int(value)
    try:
        return float(value)
    except ValueError:
        pass
    return value


def _interpolate(parent: Any, child_key: Any) -> None:
    child = parent[child_key]
    if (t := type(child)) is str:
        parent[child_key] = _get_interpolated(child)
        return
    if t is list or t is dict:
        for v in range(len(child)) if t is list else child.keys():
            _interpolate(child, v)


def _flatten(config) -> dict:
    return {k: v for k, v in _flatten_dict(config)}


def _flatten_dict(pyobj, keystring='') -> Tuple[str, Any]:
    if type(pyobj) is dict:
        for k in pyobj:
            yield from _flatten_dict(pyobj[k], str(k))
    else:
        yield keystring, pyobj



"""A setuptools based setup module.
See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
"""

import sys


# Allows override on the CLI by setting the --set-version=<version> option.
def get_version(default_version='0.5.3'):
    version_prefix = '--set-version='
    version_args = [arg for arg in sys.argv if arg.startswith(version_prefix)]
    if len(version_args) == 1:
        version_arg = version_args.pop()
        sys.argv.remove(version_arg)
        return version_arg.replace(version_prefix, '')
    if len(version_args) > 1:
        print(f'ERR: Multiple set-version arguments supplied: {version_args}.', file=sys.stderr)
        sys.exit(1)
    return default_version


VERSION = get_version()

from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='envyconfig',
    version=VERSION,
    description='YAML reader with ENV interpolation.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/geirem/envyconfig',
    author='https://github.com/geirem',
    author_email='geiremb@gmail.com',
    classifiers=[
        # https://pypi.org/classifiers/
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development',
        'License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3 :: Only',
    ],
    keywords='configtools development',
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    python_requires='>=3.8',
    extras_require={
        'test': ['pytest'],
        'googlesecrets': ["google-cloud-secret-manager"]
    },
    project_urls={  # Optional
        'Bug Reports': 'https://github.com/geirem/envyconfig/issues',
        'Funding': 'https://donate.pypi.org',
        'Source': 'https://github.com/geirem/envyconfig/',
    },
)
